#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from pandas import Series

def singleton(elems, fn=None, empty_ok=False, name='elems'):
    if isinstance(elems, Series):
        elems = elems.unique().tolist()
    if fn:
        elems = [ elem for elem in elems if fn(elem) ]
    else:
        elems = set(elems)
    if not elems:
        if empty_ok:
            return None
        raise ValueError(f'No {name} found')
    if len(elems) > 1:
        raise ValueError(f'{len(elems)} {name} found: {",".join([ str(elem) for elem in elems[:10] ])}')
    [ elem ] = elems
    return elem


# In[ ]:


from math import exp, log
from numpy import nan
from sys import stderr

def coerce(value, choices, ε=1e-2, multi_ok=False, errors='raise',warn=True):
    assert errors in ['raise','coerce','ignore',]
    elems = sorted(
        [ 
            dict(
                ratio=abs(log(value / choice)),
                choice=choice
            )
            for choice in choices
        ],
        key=lambda r: r['ratio']
    )
    log_ε = log(1+ε)
    valids = [ elem for elem in elems if elem['ratio'] <= log_ε ]
    best = elems[0]
    choice = best['choice']
    if not valids:
        msg = f'Best choice {choice} for value {value} has error {exp(best["ratio"])-1} > {ε}'
        if errors == 'raise':
            raise ValueError(msg)
        if warn:
            stderr.write(msg + '\n')
        if errors == 'coerce':
            return nan
        else:
            return value
    if len(valids) > 1 and not multi_ok:
        raise ValueError(
            '%d choices passed ε<%f filter:\n\t%s' % (
                len(valids),
                ε,
                "\n\t".join([ str(r["choice"]) for r in valids ]),
            )
        )
    return best['choice']        




#!/usr/bin/env python
# coding: utf-8

# # Context Manager utilities

# General/Common imports:

# In[ ]:


from contextlib import AbstractContextManager, contextmanager, suppress


# ## `nullcontext`
# Define `nullcontext` in a cross-version way:

# In[ ]:


try:
    # Python ≥3.7
    from contextlib import nullcontext
except ImportError:
    # Python <3.7
    class nullcontext(object):
        def __enter__(self): pass
        def __exit__(self, *args): pass


# Verify it works:

# In[2]:


with nullcontext(): pass


# ## `catch`: context manager for catching+verifying `Exception`s

# In[ ]:


class catch(AbstractContextManager):
    def __init__(self, *excs):
        self.excs = excs

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            if len(self.excs) == 1:
                raise AssertionError(f'No {self.excs[0].__name__} was thrown')
            else:
                raise AssertionError(f'None of {",".join([e.__name__ for e in self.excs])} were thrown')
        
        if not [ isinstance(exc_value, exc) for exc in self.excs ]:
            raise exc_value

        return True


# ## `no`: context manager for verifying `NameError`s (undefined variable names)

# In[1]:


no = catch(NameError)




#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from ..process import *

def exists(untracked=True, unstaged=True):
    files = lines('git','status','--porcelain')
    if not untracked:
        files = [ file for file in files if not file.startswith('??') ]
    if not unstaged:
        files = [ file for file in files if not file.startswith(' ') ]
    return bool(files)




#!/usr/bin/env python

import click
from click import command as cmd, option as opt


def paths(*args, sep=',', **kwargs):
    return opt(
        *args,
        callback=lambda ctx, param, value: [
            s
            for v in value
            for s in v.split(sep)
        ],
        multiple=True,
        **kwargs,
    )


def flag(*names, default=None, **kwargs):
    def long_opt(name):
        _default = False
        if name.startswith('no-'):
            name = name[3:]
            _default = True

        arg = f'--{name}/--no-{name}'
        return dict(
            name=name,
            default=_default,
            arg=arg,
        )

    def short_opt(name):
        if len(name) != 1:
            raise ValueError(f'Unrecognized short option name: -{name}')
        _default = name.isupper()
        name = name.lower()
        arg = f'-{name}/-{name.upper()}'
        return dict(
            name=name,
            default=_default,
            arg=arg,
        )

    def make_opt(name):
        if name.startswith('--'):
            return long_opt(name[2:])
        elif name.startswith('-'):
            return short_opt(name[1:])
        elif len(name) > 1:
            return long_opt(name)
        elif len(name) == 1:
            return short_opt(name)
        else:
            raise ValueError('Empty flag')

    opts = [ make_opt(name) for name in names ]
    if default is None:
        defaults = set([ opt['default'] for opt in opts ])
        if len(defaults) > 1: raise ValueError(f'Conflicting defaults: {opts}')
        [default] = defaults

    return opt(
        *[ opt['arg'] for opt in opts ],
        default=default,
        **kwargs
    )



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

from utz import *
rgx = r'v?(?P<version>\d+\.\d+\.\d+)'
tag = line('git','describe','--tags','HEAD')
version = match(rgx, tag)['version']

setup(
    name="utz",
    version=version,
    author="Ryan Williams",
    author_email="ryan@runsascoded.com",
    description="Misc stdlib, pandas, subprocess, and other utilities, exposed for easy importing + boilerplate-reduction",
    install_requires=[
        "GitPython",
        "joblib",
        "pandas",
        "pyyaml",
    ],
    extras_require={
        'pdf': [
            'reportlab==3.5.42',
            'pyPDF2==1.26.0',
        ],
    },
    scripts=['utz.sh',],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/runsascoded/utz",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



#!/usr/bin/env python
# coding: utf-8

from inspect import getfullargspec
import json
from jupyter_client import kernelspec
from os import makedirs, remove
from os.path import abspath, basename, dirname, exists, join, splitext
from papermill import execute_notebook
from pathlib import Path
from sys import executable
from traceback import format_exception

from .collections import singleton
from . import git
from .process import line, run
from papermill import PapermillExecutionError


def current_kernel():
    kernels = kernelspec.find_kernel_specs()
    kernel = singleton(
        [ 
            name
            for name, kernel_dir
            in kernels.items()
            if json.load(open(f'{kernel_dir}/kernel.json','r'))['argv'][0] == executable
        ], empty_ok=True
    )
    if kernel: return kernel
    return singleton(kernels.keys())


def execute(
    input,
    output=None,
    # Papermill execute_notebook kwargs that we may override defaults for
    nest_asyncio=True,
    cwd=False,
    inject_paths=False,
    progress_bar=False,
    # Aliases for papermill kwargs
    kernel=None,
    params=None,
    # Configs for committing run notebook + specifying output paths to include
    commit=True,
    msg=None,
    start_sha=None,
    msg_path='_MSG',
    *args,
    **kwargs
):
    '''Run a jupyter notebook using papermill, and git commit the output
    '''
    if not exists(input) and not input.endswith('.ipynb'):
        input += '.ipynb'
    if not exists(input):
        raise ValueError(f"Nonexistent input notebook: {input}")
    if commit:
        if not start_sha:
            start_sha = git.head.sha()
    name = splitext(input)[0]
    if output:
        if not output.endswith('.ipynb'):
            output = join(output, basename(input))
        out_dir = dirname(abspath(output))
        makedirs(out_dir, exist_ok=True)
    else:
        output = input

    # Parse execute_notebook() kwargs names
    spec = getfullargspec(execute_notebook)
    exec_kwarg_names = spec.args[-len(spec.defaults):]

    # Separate out execute_notebook kwargs
    exec_kwargs = { 
        name: kwargs.pop(name)
        for name in exec_kwarg_names
        if name in kwargs
    }

    # Merge "kernel" and "kernel_name" kwargs; default to current kernel
    if kernel:
        if 'kernel_name' in exec_kwargs:
            if exec_kwargs['kernel_name'] != kernel:
                raise ValueError(f'Conflicting kernel_name values: {exec_kwargs["kernel_name"]} vs. {kernel}')
        else:
            exec_kwargs['kernel_name'] = kernel
    else:
        if 'kernel_name' not in exec_kwargs:
            kernel_name = current_kernel()
            exec_kwargs['kernel_name'] = kernel_name

    # Support "params" and "parameters" aliases for execute_notebooks' "parameters" kwarg; pick whichever one is present here (and merge them if they both are, raising on conflicting values for the same key)
    if params is not None:
        if 'parameters' in exec_kwargs:
            parameters = exec_kwargs['parameters']
            for k,v in params.items():
                if k in parameters and parameters[k] != v:
                    raise ValueError(
                        f'Conflicting kwargs: params %s vs parameters %s' % (
                            str(v),
                            str(parameters[k])
                        )
                    )
            if exec_kwargs['kernel_name'] != kernel:
                raise ValueError(f'Conflicting kernel_name values: {exec_kwargs["kernel_name"]} vs. {kernel}')
        else:
            exec_kwargs['parameters'] = params

    parameters = exec_kwargs.get('parameters')
    if parameters:
        if kwargs:
            raise ValueError(f'Passing `parameters` arg to papermill, but found dangling kwargs: {kwargs}')
    else:
        parameters = kwargs
        exec_kwargs['parameters'] = parameters

    # Convert Path objects to strings, as a courtesy/convenience (they will fail to serialize in execute_notebook() otherwise)
    exec_kwargs['parameters'] = {
        k: str(v) if isinstance(v, Path) else v
        for k,v in exec_kwargs['parameters'].items()
    }

    # We take "cwd=True" to mean "run in the current directory" (papermill's default); our default is otherwise to run papermill from the directory containing the notebook being executed
    if cwd:
        if cwd is True:
            cwd = None
    else:
        cwd = dirname(abspath(input))

    exc = None
    try:
        execute_notebook(
            str(input),
            str(output),
            *args,
            nest_asyncio=nest_asyncio,  # allow papermill-in-papermill
            cwd=cwd,
            inject_paths=inject_paths,  # normally unused, but allow notebook to reflect on its own path
            progress_bar=progress_bar,
            **exec_kwargs,
        )
    except PapermillExecutionError as e:
        # Allow notebooks to short-circuit execution by raising an Exception whose message begins with the string "OK: "
        if e.evalue.startswith('OK: '):
            print('Run notebook %s exited with "OK" msg' % str(input))
        else:
            exc = e

    if commit or exc:
        if exc:
            msg = '\n'.join(
                [
                    f'Failed: {repr(exc)}',
                    '',
                    print(''.join(
                        format_exception(
                            etype=type(exc),
                            value=exc,
                            tb=exc.__traceback__,
                        )
                    )),
                ]
            )
        # Commit results:
        # - by default, just the notebook output path
        # - pass a list of paths to "commit" (or a single path as a str or Path) to include them in the commit
        # - if a file named '_MSG' is written by the notebook, use its contents as the commit message
        if commit is True:
            commit = []
        elif isinstance(commit, str):
            commit = [commit]
        elif isinstance(commit, Path):
            commit = [str(commit)]
        commit += [output]
        if not msg:
            if exists(msg_path):
                with open(msg_path,'r') as f:
                    msg = f.read()
                remove(msg_path)
            else:
                msg = name
        last_sha = git.head.sha()
        run(['git','add'] + commit)
        run('git','commit','-m',msg)
        if start_sha != last_sha:
            repo = git.Repo()
            tree = repo.tree().hexsha
            head = line('git','commit-tree',tree,'-p',start_sha,'-p',last_sha,'-m',msg)
            run('git','reset',head)

        if exc:
            raise exc
